
#include <Y/main/control.h>
#include <Y/util/yutil.h>
#include <Y/util/index.h>
#include <Y/util/pqueue.h>
#include <stdio.h>
#include <sys/time.h>

static int controlRunning = 1;

struct ControlFileDescriptor
{
  int fd;
  int watchMask;
  void *userData;
  void (*callback)(int, int, void *);
};

static int
controlFileDescriptorsKeyFunction (void *key_v, void *obj_v)
{
  int *key_p = key_v;
  struct ControlFileDescriptor *obj = (struct ControlFileDescriptor *)obj_v;
  return *key_p - obj -> fd;  
}

static int
controlFileDescriptorsComparisonFunction (void *obj1_v, void *obj2_v)
{
  struct ControlFileDescriptor *obj1 = (struct ControlFileDescriptor *)obj1_v;
  struct ControlFileDescriptor *obj2 = (struct ControlFileDescriptor *)obj2_v;
  return obj1->fd - obj2->fd;
}

static void
controlFileDescriptorsDestructorFunction (void *obj_v)
{
  free (obj_v);
}

struct ControlTimedEvent
{
  struct timeval time;
  int id;
  void *userData;
  void (*callback)(void *);
};

static int nextTimedEventID = 1;

static int
controlTimedEventsComparisonFunction (void *obj1_v, void *obj2_v)
{
  struct ControlTimedEvent *obj1 = obj1_v;
  struct ControlTimedEvent *obj2 = obj2_v;
  if (obj1 -> time.tv_sec == obj2 -> time.tv_sec)
    {
       return obj1 -> time.tv_usec - obj2 -> time.tv_usec;
    }
  else
    {
       return obj1 -> time.tv_sec - obj2 -> time.tv_sec;
    }
}

static void
controlTimedEventDestructorFunction (void *obj_v)
{
  yfree (obj_v);
}

static struct Index *fileDescriptors; 
static struct PQueue *timedEvents; 

void
controlInitialise ()
{
  fileDescriptors = indexCreate (controlFileDescriptorsKeyFunction,
                                 controlFileDescriptorsComparisonFunction);
  timedEvents = pqueueCreate (controlTimedEventsComparisonFunction);
}

void
controlRegisterFileDescriptor (int fd, int watchMask, void *userData,
              void (*callback)(int fd, int causeMask, void *userData))
{
  struct ControlFileDescriptor *obj =
    ymalloc (sizeof (struct ControlFileDescriptor));
  obj -> fd = fd;
  obj -> watchMask = watchMask;
  obj -> userData = userData;
  obj -> callback = callback;
  indexAdd (fileDescriptors, obj);
}

void
controlChangeFileDescriptorMask (int fd, int watchMask)
{
  struct ControlFileDescriptor *obj = indexFind (fileDescriptors, &fd);
  if (obj != NULL)
    {
      obj -> watchMask = watchMask;
    }
}

void
controlUnregisterFileDescriptor (int fd)
{
  void *obj = indexRemove (fileDescriptors, &fd);
  free (obj);
}

int
controlTimerDelay (int minIntervalSeconds, int minIntervalMilliseconds,
                   void *userData, void (*callback)(void *userData))
{
  struct timeval now;
  struct ControlTimedEvent *event;

  event = ymalloc (sizeof (struct ControlTimedEvent));
  event -> userData = userData;
  event -> callback = callback;
  event -> id = nextTimedEventID ++;

  gettimeofday (&now, NULL);

  event -> time.tv_usec = (minIntervalMilliseconds * 1000 + now.tv_usec);
  event -> time.tv_sec = minIntervalSeconds + now.tv_sec;
  if (event -> time.tv_usec >= 1000000)
    {
      event -> time.tv_sec++;
      event -> time.tv_usec -= 1000000;
    }

  pqueueInsert (timedEvents, event); 

  return event -> id;
}

static int
controlCancelTimerDelayTestFunction (void *timedEvent_v, void *id_v)
{
  int *id = id_v;
  struct ControlTimedEvent *timedEvent = timedEvent_v;
  return (timedEvent -> id == *id);
}

void
controlCancelTimerDelay (int id)
{
  pqueueRemove (timedEvents, &id, controlCancelTimerDelayTestFunction);  
}

static int
controlIteration ()
{
  struct ControlFileDescriptor *cfd;
  struct ControlTimedEvent *tev;
  struct IndexIterator *iter;
  struct timeval timeout;
  int maxFd = 0;
  int retval;
  fd_set fds[3];
  FD_ZERO (&fds[0]); 
  FD_ZERO (&fds[1]);  
  FD_ZERO (&fds[2]);

  iter = indexGetStartIterator (fileDescriptors);
  while (indexiteratorHasValue (iter))
    {
      cfd = indexiteratorGet (iter);
      if (cfd -> watchMask & CONTROL_WATCH_READ)
        FD_SET (cfd -> fd, &fds[0]);
      if (cfd -> watchMask & CONTROL_WATCH_WRITE)
        FD_SET (cfd -> fd, &fds[1]);
      if (cfd -> watchMask & CONTROL_WATCH_EXCEPT)
        FD_SET (cfd -> fd, &fds[2]);
      indexiteratorNext (iter);
    }

  /* need to find maximum file descriptor */
  indexiteratorPrevious (iter); /* back up one place */
  cfd = indexiteratorGet (iter);
  if (cfd != NULL)
    maxFd = cfd -> fd;
  indexiteratorDestroy (iter);

  /* work out time to next poll or timer event */
  tev = pqueuePeekNext (timedEvents);
  if (tev)
    {
      struct timeval now;
      gettimeofday (&now, NULL);
      timeout.tv_sec = tev -> time.tv_sec - now.tv_sec;
      timeout.tv_usec = tev -> time.tv_usec - now.tv_usec;
      if (timeout.tv_usec < 0)
        {
          timeout.tv_sec--;
          timeout.tv_usec += 1000000;
        }
    }
  else
    {
      timeout.tv_sec = 5;
      timeout.tv_usec = 0;
    }

  /* select () */
  if (timeout.tv_sec < 0 || (timeout.tv_sec == 0 && timeout.tv_usec == 0))
    {
      timeout.tv_sec = 0; timeout.tv_usec = 100; 
    }

  retval = select (maxFd + 1, &fds[0], &fds[1], &fds[2], &timeout);

  /* despatch whatever woke us up */
  if (retval > 0)
    {
      int i = 0;
      for (i = 0; i <= maxFd; ++i)
        {
          int mask = 0;
          if (FD_ISSET (i, &fds[0]))
            mask |= CONTROL_WATCH_READ;
          if (FD_ISSET (i, &fds[1]))
            mask |= CONTROL_WATCH_WRITE;
          if (FD_ISSET (i, &fds[2]))
            mask |= CONTROL_WATCH_EXCEPT;
          if (mask)
            {
              cfd = indexFind (fileDescriptors, &i);
              if (cfd)
                cfd -> callback (i, mask, cfd -> userData);
            }
        }
    }

   /* call whatever timers have expired */
   if ((tev = pqueuePeekNext (timedEvents)) != NULL)
     {
       struct timeval now;
       gettimeofday (&now, NULL);
       while (now.tv_sec > tev -> time.tv_sec
              || (now.tv_sec == tev -> time.tv_sec
                  && now.tv_usec > tev -> time.tv_usec))
         {
           tev = pqueueGetNext (timedEvents);
           tev -> callback (tev -> userData);
           yfree (tev);
           tev = pqueuePeekNext (timedEvents);
           if (tev == NULL)
             break;
         }
     }

  return controlRunning;
}

void
controlRun ()
{
  while (controlIteration ())
    ;
}

void
controlShutdownY ()
{
  controlRunning = 0;
}

void
controlFinalise ()
{
  indexDestroy (fileDescriptors, controlFileDescriptorsDestructorFunction);
  pqueueDestroy (timedEvents, controlTimedEventDestructorFunction);
}

/* arch-tag: 902f698b-690f-44a7-a931-588a000282db
 */
