#include <Y/message/message.h>
#include <Y/message/client.h>
#include <Y/message/special.h>
#include <Y/const.h>
#include <Y/object/class.h>
#include <Y/object/object.h>
#include <Y/util/yutil.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

struct Message *messageCreate (int to, int from, int cid, int oid,
                               int op, int meta, int ext, int len, char *data)
{
  struct Message *m = ymalloc (sizeof (struct Message));
  m -> to   = to;
  m -> from = from;
  m -> cid  = cid;
  m -> oid  = oid;
  m -> op   = op;
  m -> meta = meta;
  m -> ext  = ext;
  m -> len  = len;
  m -> data = data;
  return m;
}

void
messageDestroy (struct Message *m)
{
  if (m != NULL)
    {
      yfree (m->data);
      yfree (m);
    }
}

static void
messageLog (struct Message *m)
{
}

static void
messageDespatchNoOperation (struct Client *clientFrom, struct Message *m)
{
  /* NO_OP */ 
  messageDestroy (m);
}

static void
messageDespatchIgnore (struct Client *clientFrom, struct Message *m)
{
  Y_WARN ("Ignoring message.");
  messageDestroy (m);
}

static void
messageDespatchInvokeSpecial (struct Client *clientFrom, struct Message *m)
{
  void (*special)(struct Parameters *);
  int nameLen = strcspn(m->data, "\x1c");
  struct Parameters *params = parametersBuild (m->data + nameLen + 1,
                                               m->len - nameLen - 1);
  char *r = NULL;
  int rl = 0;
  m->data[nameLen] = '\0';
  special = specialFind (m->data);
  if (special != NULL)
    special (params);
  else
    parametersAddOutput (params, "Not found.");
  r = parametersGetOutputData (params, NULL, &rl);
  if ((m->meta & 1<<0) != 0)
    messageDespatch ( NULL,
        messageCreate (clientGetID (clientFrom), 0, 0, 0,
                       YMO_SPECIAL_RETURNS, 1, 0, rl, r));
  parametersDestroy (params);
  messageDestroy (m);
}

static void
messageDespatchFindClass (struct Client *clientFrom, struct Message *m)
{
  struct Class *class = classFindByName (m->data);
  struct Message *reply;
  if (class != NULL)
    reply = messageCreate (clientGetID (clientFrom), 0, classGetID (class), 0,
                           YMO_FOUND_CLASS, 0, 0, 0, NULL);
  else
    reply = messageCreate (clientGetID (clientFrom), 0, 0, 0,
                           YMO_ERROR, 0, 0, 17, ystrdup ("Class not found."));
  messageDespatch (NULL, reply);
  messageDestroy (m);
}

static void
messageDespatchInstantiate (struct Client *clientFrom, struct Message *m)
{
  struct Class *class = classFindByID (m->cid);
  struct Message *reply;
  if (class != NULL)
    {
      struct Parameters *params;
      struct Object *object;

      params = parametersBuild (m->data, m->len);
      object = classInstantiate (class, clientFrom, params);

      if (object != NULL)
        {
          int repDataLen;
          char *repData = parametersGetOutputData (params, NULL, &repDataLen);
          reply = messageCreate (clientGetID (clientFrom), 0,
                                 classGetID (class), objectGetID (object),
                                 YMO_NEW_OBJECT, 0, 0, repDataLen, repData);
        }
      else
        {
          reply = messageCreate (clientGetID (clientFrom), 0,
                                 classGetID (class), 0,
                                 YMO_ERROR, 0, 0, 29,
                                 ystrdup ("Could not instantiate class."));
        }
      parametersDestroy (params);
    }
  else
    {
      reply = messageCreate (clientGetID (clientFrom), 0, 0, 0,
                             YMO_ERROR, 0, 0, 17, ystrdup ("Class not found."));
    }
  messageDespatch (NULL, reply);
  messageDestroy (m);
}

static void
messageDespatchInvokeMethod (struct Client *clientFrom, struct Message *m)
{
  struct Class *class = classFindByID (m->cid);
  if (class != NULL)
    {
      struct Object *object = objectFind (m->oid);
      int nameLen = strcspn(m->data, "\x1c");
      struct Parameters *params = parametersBuild (m->data + nameLen + 1,
                                                   m->len - nameLen - 1);
      char *r = NULL;
      int rl = 0;
      m->data[nameLen] = '\0';
      classInvokeMethod (class, object, m->data, params);
      r = parametersGetOutputData (params, NULL, &rl);
      if ((m->meta & 1<<0) != 0)
        messageDespatch ( NULL,
            messageCreate (clientGetID (clientFrom), 0, m->cid, m->oid,
                           YMO_METHOD_RETURNS, 1, 0, rl, r));
      parametersDestroy (params);
    }
  messageDestroy (m);
}

static void
messageDespatchGetProperty (struct Client *clientFrom, struct Message *m)
{
  messageDestroy (m);
}

static void
messageDespatchSetProperty (struct Client *clientFrom, struct Message *m)
{
  struct Class *class = classFindByID (m->cid);
  if (class != NULL)
    {
      struct Object *object = objectFind (m->oid);
      if (object != NULL)
        {
          int nameLen = strcspn(m->data, "\x1c");
          m->data[nameLen] = '\0';
          objectSetPropertyToString (object, m->data, m->data+nameLen + 1);
        }
    }
  messageDestroy (m);
}

static void
messageDespatchSubscribeSignal (struct Client *clientFrom, struct Message *m)
{
  struct Object *object = objectFind (m->oid);
  if (object != NULL)
    {
      int nameLen = strcspn(m->data, "\x1c");
      m->data[nameLen] = '\0';
      objectSubscribeSignal (object, m->data);
    }
  messageDestroy (m);
}

static void
messageDespatchListMethods (struct Client *clientFrom, struct Message *m)
{
  messageDestroy (m);
}

static void
messageDespatchListProperties (struct Client *clientFrom, struct Message *m)
{
  messageDestroy (m);
}


void
messageDespatch (struct Client *clientFrom, struct Message *m)
{
  struct Class *class = NULL;
  struct Object *object = NULL;
  struct Client *clientTo = NULL;

  messageLog (m);

  /* check the message has a valid source. */
  if (clientFrom == NULL && m->from != 0)
    {
      clientFrom = clientFind (m->from);
      if (clientFrom == NULL)
        {
          Y_WARN ("Unknown message source (%d), ignoring.", m->from);
          messageDestroy (m);
          return;
        }
    }

  /* is the message addressed to another client? */
  if (m->to != 0)
    {
      clientTo = clientFind (m->to);
      if (clientTo == NULL)
        {
          Y_WARN ("Unknown message destination (%d), ignoring.", m->to);
          messageDestroy (m);
          return;
        }
      clientSendMessage (clientTo, m);
      return;
    }

  /* despatch standard operations */
  switch (m->op)
    {
      case YMO_NO_OPERATION:     messageDespatchNoOperation (clientFrom, m);
                                 return;
      case YMO_INVOKE_SPECIAL:   messageDespatchInvokeSpecial (clientFrom, m);
                                 return;
      case YMO_FIND_CLASS:       messageDespatchFindClass (clientFrom, m);
                                 return;
      case YMO_INSTANTIATE:      messageDespatchInstantiate (clientFrom, m);
                                 return;
      case YMO_INVOKE_METHOD:    messageDespatchInvokeMethod (clientFrom, m);
                                 return;
      case YMO_GET_PROPERTY:     messageDespatchGetProperty (clientFrom, m);
                                 return;
      case YMO_SET_PROPERTY:     messageDespatchSetProperty (clientFrom, m);
                                 return;
      case YMO_SUBSCRIBE_SIGNAL: messageDespatchSubscribeSignal (clientFrom, m);
                                 return;
      case YMO_LIST_METHODS:     messageDespatchListMethods (clientFrom, m);
                                 return;
      case YMO_LIST_PROPERTIES:  messageDespatchListProperties (clientFrom, m);
                                 return;
      case YMO_ERROR:            clientClose (clientFrom);
                                 messageDestroy (m);
                                 return;
      case YMO_EVENT:
      case YMO_SPECIAL_RETURNS:
      case YMO_FOUND_CLASS:
      case YMO_NEW_OBJECT:
      case YMO_METHOD_RETURNS:     /* these should not be received by Y, */
      case YMO_GOT_PROPERTY:       /* ignore them. */
      case YMO_METHODS_LISTED:
      case YMO_PROPERTIES_LISTED: messageDespatchIgnore (clientFrom, m);
                                  return; 

      default: ;
    }

  /* pass other messages onto their class. */
  if (m->cid != 0)
    {
      class = classFindByID (m->cid);
      if (class == NULL)
        {
          Y_WARN ("Unknown message class (%d), ignoring.", m->cid);
          messageDestroy (m);
          return;
        }
      classDespatch (class, object, m);
      messageDestroy (m);
      return;
    }

  Y_WARN ("Message not handled.");
  messageDestroy (m);
}
 

/* arch-tag: 2edc8cfa-2fdd-45ea-851b-db14440d3205
 */
