#ifndef Y_SCREEN_RENDERER_H
#define Y_SCREEN_RENDERER_H

struct Renderer;

#include <Y/y.h>
#include <Y/util/rectangle.h>
#include <stdint.h>

/* This represents an ABSTRACT renderer.
 *
 * A Renderer is a "visitor" in the Visitor pattern.  A specific
 * implementation if visitor is selected by the video driver, and
 * is then passed over the Widget hierarchy.  The renderer provides
 * the interface by which Buffer-to-Video transfers may occur.
 */

/* Signal that all processing for this renderer is complete. Renderers
 * that buffer information should flush the buffer here.
 */
void rendererComplete (struct Renderer *);

/* Destroy the renderer.  If the renderer has not been "Complete"d, then
 * the state of the display for the renderer's rectangle is undefined.
 */
void rendererDestroy (struct Renderer *);

/* Attempt to enter the region (X, Y, W, H).
 * Translate all further coordinates by DX and DY (until the corresponding
 * Leave).
 * Entering succeeds if there is a non-zero resulting region.
 *
 * Returns 1 if entering succeeds.
 */
int rendererEnter (struct Renderer *, const struct Rectangle *rect,
                   int dx, int dy);

/* Leave the last region enterered.
 */
void rendererLeave (struct Renderer *);

/* Blit (blending) a rectangle of RGBA data.
 * 
 * The data should be blitted to the region-based co-ordinates (X,Y).
 * The data starts at the position given by DATA. The data consists of rows
 * of W pixels worth of RGBA colour data, with each row S pixels from the
 * previous. H rows should be blitted.
 */
void rendererBlitRGBAData (struct Renderer *, int x, int y, uint32_t *data,
                           int w, int h, int s);

/* Draw a rectangle filled with a solid colour, COLOUR. The rectangle should
 * placed with the upper-left corner at the device independent co-ordinates
 * (X,Y), and be W by H in size.
 */
void rendererDrawFilledRectangle (struct Renderer *, uint32_t colour,
                                  int x, int y, int w, int h);

#endif /* Y_BUFFER_RENDERER_H */

/* arch-tag: 0888bf20-161a-421c-9257-1ee7c46c64cf
 */
