#include <Y/widget/button.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/buffer/painter.h>

#include <Y/modules/theme.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>

#include <Y/text/font.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>

struct Button
{
  struct Widget widget;
};

static void buttonReconfigure (struct Widget *);
static void buttonPaint (struct Widget *, struct Painter *);
static int buttonPointerButton (struct Widget *, int, int, int, int);
static int buttonPointerMotion (struct Widget *, int, int, int, int);
static void buttonPointerEnter (struct Widget *, int, int);
static void buttonPointerLeave (struct Widget *);

static struct Object *buttonInstantiate (struct Parameters *);

static void buttonTextSet (struct Property *, void *);

CLASSDEF struct Class buttonClass =
{
  name:          "Button",
  super:         &widgetClass,
  instantiate:   buttonInstantiate,
  destroy:       (void(*)(struct Object *))buttonDestroy,
  methodCount:   0,
  methods:
    { {} }
};

static struct WidgetTable buttonTable =
{
  c:             &buttonClass,
  reconfigure:   buttonReconfigure,
  paint:         buttonPaint,
  pointerButton: buttonPointerButton,
  pointerMotion: buttonPointerMotion,
  pointerEnter:  buttonPointerEnter,
  pointerLeave:  buttonPointerLeave
};

static inline struct Button *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == buttonClass ); */
  return (struct Button *)widget;
}

static inline const struct Button *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == buttonClass ); */
  return (const struct Button *)widget;
}

static void
buttonPaint (struct Widget *self_w, struct Painter *painter)
{
  struct Font *font;
  const char *text =
               objectGetPropertyAsString (widgetToObject (self_w), "text");

  painterSaveState (painter);
  themeDrawButtonPane (painter, 0, 0,
                                self_w -> w, self_w -> h, self_w -> state);
  painterSetPenColour (painter, 0xFF000000);
  int offset, width, ascender, descender;
  font = themeGetDefaultFont ();
  fontGetMetrics (font, &ascender, &descender, NULL);
  fontMeasureString (font, text, &offset, &width, NULL);
  fontRenderString (font, painter, text, (self_w -> w - width + offset) / 2 - offset,
                                         (self_w -> h - ascender - descender) / 2 + ascender);
  painterRestoreState (painter);
}

struct Button *
buttonCreate ()
{
  struct Button *self = ymalloc (sizeof (struct Button));
  objectInitialise (&(self -> widget.o), &buttonClass);
  widgetInitialise (&(self -> widget), &buttonTable);
  objectPropertyCallOnSet (buttonToObject (self), "text", self, buttonTextSet);
  return self;
}

void
buttonDestroy (struct Button *self)
{
  widgetFinalise (buttonToWidget (self));
  objectFinalise (buttonToObject (self));
  yfree (self);
}

struct Widget *
buttonToWidget (struct Button *self)
{
  return &(self -> widget);
}

struct Object *
buttonToObject (struct Button *self)
{
  return &(self -> widget.o);
}

static struct Object *
buttonInstantiate (struct Parameters *p)
{
  return buttonToObject (buttonCreate ());
}

static void buttonReconfigure (struct Widget *self_w)
{
  struct Button *self = castBack (self_w);
  struct Font *font;
  const char *text =
               objectGetPropertyAsString (widgetToObject (self_w), "text");
  int offset, width, ascender, descender;
  font = themeGetDefaultFont ();
  fontGetMetrics (font, &ascender, &descender, NULL);
  fontMeasureString (font, text, &offset, &width, NULL);

  self -> widget.minWidth = width - offset + 10;
  self -> widget.minHeight = ascender + descender + 12;
  self -> widget.reqWidth = width - offset + 10;
  self -> widget.reqHeight =  ascender + descender + 12;

  /* we would prefer to be square rather than narrow... */
  if (self -> widget.reqWidth < self -> widget.reqHeight)
    self -> widget.reqWidth = self -> widget.reqHeight;

  widgetReconfigure (self -> widget.container);
}

static void
buttonTextSet (struct Property *text, void *self_v)
{
  struct Button *self = self_v;
  widgetReconfigure (buttonToWidget (self));
  widgetRepaint (buttonToWidget (self), NULL);
}

static int
buttonPointerButton (struct Widget *self_w, int x, int y, int b, int pressed)
{
  struct Button *self = castBack (self_w);
  if (b==0)
    {
      if (pressed)
        {
          pointerGrab (self_w);
          self -> widget.state = WIDGET_STATE_PRESSED;
          widgetRepaint (self_w, NULL); 
        }
      else
        {
          pointerRelease ();
          if (self -> widget.state == WIDGET_STATE_PRESSED)
            {
              objectEmitSignal (buttonToObject (self), "clicked", NULL);
            }
          if (widgetContainsPoint (self_w, x, y))
            self -> widget.state = WIDGET_STATE_HOVER;
          else
            self -> widget.state = WIDGET_STATE_NORMAL;
          widgetRepaint (self_w, NULL);
        }
    }
  return 1;
}

static int
buttonPointerMotion (struct Widget *self_w, int x, int y, int dx, int dy)
{
  struct Button *self = castBack (self_w);
  if (self -> widget.state == WIDGET_STATE_PRESSED
      && !widgetContainsPoint (self_w, x, y))
    {
      self -> widget.state = WIDGET_STATE_CANCELLING;
      widgetRepaint (self_w, NULL);
    }
  else if (self -> widget.state == WIDGET_STATE_CANCELLING
      && widgetContainsPoint (self_w, x, y))
    {
      self -> widget.state = WIDGET_STATE_PRESSED;
      widgetRepaint (self_w, NULL);
    }
  return 1;
}

static void
buttonPointerEnter (struct Widget *self_w, int x, int y)
{
  struct Button *self = castBack (self_w);
  self -> widget.state = WIDGET_STATE_HOVER;
  widgetRepaint (buttonToWidget (self), NULL);
}

static void
buttonPointerLeave (struct Widget *self_w)
{
  struct Button *self = castBack (self_w);
  self -> widget.state = WIDGET_STATE_NORMAL;
  widgetRepaint (buttonToWidget (self), NULL);
}


/* arch-tag: 791d4e69-40ee-48d3-885f-fcb076464847
 */
