#include <Y/widget/canvas.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/buffer/painter.h>
#include <Y/buffer/rgbabuffer.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>

#include <Y/text/font.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>

struct Canvas
{
  struct Widget widget;
  struct RGBABuffer *front, *back;
  struct Painter *painter;
  int resizing;
};

static void canvasResize (struct Widget *);
static void canvasRender (struct Widget *, struct Renderer *);
static void canvasDespatch (struct Canvas *, struct Message *);

static struct Object *canvasInstantiate (struct Parameters *);

static void canvasCSavePainterState    (struct Canvas *, struct Parameters *);
static void canvasCRestorePainterState (struct Canvas *, struct Parameters *);
static void canvasCSetBlendMode        (struct Canvas *, struct Parameters *);
static void canvasCSetPenColour        (struct Canvas *, struct Parameters *);
static void canvasCSetFillColour       (struct Canvas *, struct Parameters *);
static void canvasCReset               (struct Canvas *, struct Parameters *);
static void canvasCClearRectangles     (struct Canvas *, struct Parameters *);
static void canvasCDrawRectangles      (struct Canvas *, struct Parameters *);
static void canvasCDrawHLines          (struct Canvas *, struct Parameters *);
static void canvasCDrawVLines          (struct Canvas *, struct Parameters *);
static void canvasCDrawLines           (struct Canvas *, struct Parameters *);
static void canvasCSwapBuffers         (struct Canvas *, struct Parameters *);
static void canvasCRequestSize         (struct Canvas *, struct Parameters *);

CLASSDEF struct Class canvasClass =
{
  name:          "Canvas",
  super:         &widgetClass,
  instantiate:   canvasInstantiate,
  destroy:       (void (*)(struct Object *))canvasDestroy,
  despatch:      (void (*)(struct Object *, struct Message *))canvasDespatch,
  methodCount:   13,
  methods:
  {
    { "clearRectangles",     (ObjectMethod_t) canvasCClearRectangles     },
    { "drawHLines",          (ObjectMethod_t) canvasCDrawHLines          },
    { "drawLines",           (ObjectMethod_t) canvasCDrawLines           },
    { "drawRectangles",      (ObjectMethod_t) canvasCDrawRectangles      },
    { "drawVLines",          (ObjectMethod_t) canvasCDrawVLines          },
    { "requestSize",         (ObjectMethod_t) canvasCRequestSize         },
    { "reset",               (ObjectMethod_t) canvasCReset               },
    { "restorePainterState", (ObjectMethod_t) canvasCRestorePainterState },
    { "savePainterState",    (ObjectMethod_t) canvasCSavePainterState    },
    { "setBlendMode",        (ObjectMethod_t) canvasCSetBlendMode        },
    { "setFillColour",       (ObjectMethod_t) canvasCSetFillColour       },
    { "setPenColour",        (ObjectMethod_t) canvasCSetPenColour        },
    { "swapBuffers",         (ObjectMethod_t) canvasCSwapBuffers         }
  }
};

static struct WidgetTable canvasTable =
{
  c:             &canvasClass,
  resize:        canvasResize,
  render:        canvasRender,
};

static inline struct Canvas *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == canvasClass ); */
  return (struct Canvas *)widget;
}

static inline const struct Canvas *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == canvasClass ); */
  return (const struct Canvas *)widget;
}

static void
canvasRender (struct Widget *self_w, struct Renderer *renderer)
{
  struct Canvas *self = castBack (self_w);
  int fw, fh, bw, bh;
  rgbabufferGetSize (self -> front, &fw, &fh);
  rgbabufferGetSize (self -> back, &bw, &bh); 
  rgbabufferRender (self -> front, renderer, 0, 0);
}

struct Canvas *
canvasCreate ()
{
  struct Canvas *self = ymalloc (sizeof (struct Canvas));
  struct Rectangle rect;
  objectInitialise (&(self -> widget.o), &canvasClass);
  widgetInitialise (&(self -> widget), &canvasTable);
  self -> front = rgbabufferCreate ();
  self -> back  = rgbabufferCreate ();
  self -> resizing = 0;
  rgbabufferSetSize (self -> front, 64, 64);
  rgbabufferSetSize (self -> back, 64, 64);
  self -> painter = rgbabufferGetPainter (self -> back);
  rect.x = 0;
  rect.y = 0;
  rect.w = 64;
  rect.h = 64;
  painterClipTo (self -> painter, &rect);
  painterSetFillColour (self -> painter, 0);
  painterClearRectangle (self -> painter, 0, 0, 64, 64);
  return self;
}

void
canvasDestroy (struct Canvas *self)
{
  painterDestroy (self -> painter);
  rgbabufferDestroy (self -> front);
  rgbabufferDestroy (self -> back);
  widgetFinalise (canvasToWidget (self));
  objectFinalise (canvasToObject (self));
  yfree (self);
}

struct Widget *
canvasToWidget (struct Canvas *self)
{
  return &(self -> widget);
}

struct Object *
canvasToObject (struct Canvas *self)
{
  return &(self -> widget.o);
}

static struct Object *
canvasInstantiate (struct Parameters *p)
{
  return canvasToObject (canvasCreate ());
}

static void
canvasDespatch (struct Canvas *self, struct Message *message)
{
  Y_TRACE ("Received Message: op:%4d    meta:%4d    ext:%4d",
           message->op, message->meta, message->ext); 
}

void
canvasResize (struct Widget *self_w)
{
  struct Canvas *self = castBack (self_w);

  if (self -> resizing == 0)
   {
     objectEmitSignal (canvasToObject (self), "resize", NULL);
     self -> resizing = 1;
   }
  
}

void
canvasCSavePainterState (struct Canvas *self, struct Parameters *params)
{
  painterSaveState (self -> painter);
}

void
canvasCRestorePainterState (struct Canvas *self, struct Parameters *params)
{
  painterRestoreState (self -> painter);
}

void
canvasCSetBlendMode (struct Canvas *self, struct Parameters *params)
{
}

void
canvasCSetPenColour (struct Canvas *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 1)
    {
      painterSetPenColour (self -> painter,
                           strtoul (parametersGetInput (params, 0), NULL, 0));
    }
}

void
canvasCSetFillColour (struct Canvas *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 1)
    {
      painterSetFillColour (self -> painter,
                            strtoul (parametersGetInput (params, 0), NULL, 0));
    }
}

void
canvasCReset (struct Canvas *self, struct Parameters *params)
{
  uint32_t fc = painterGetFillColour (self -> painter);
  self -> resizing = 0;
  painterSetFillColour (self -> painter,
    objectGetPropertyAsInteger (canvasToObject (self), "background"));
  rgbabufferSetSize (self -> back, self -> widget.w, self -> widget.h);
  painterDestroy (self -> painter);
  self -> painter = rgbabufferGetPainter (self -> back);
  painterClearRectangle (self -> painter, 0, 0,
                         self -> widget.w, self -> widget.h);
  painterSetFillColour (self -> painter, fc);
  parametersAddOutputInteger (params, self -> widget.w);
  parametersAddOutputInteger (params, self -> widget.h);
}

void
canvasCClearRectangles (struct Canvas *self, struct Parameters *params)
{
}

void
canvasCDrawRectangles (struct Canvas *self, struct Parameters *params)
{
}

void
canvasCDrawHLines (struct Canvas *self, struct Parameters *params)
{
}

void
canvasCDrawVLines (struct Canvas *self, struct Parameters *params)
{
}

void
canvasCDrawLines (struct Canvas *self, struct Parameters *params)
{
  int i, n = parametersGetInputCount (params);
  for (i=0; i<n-3; i+=4)
    {
      int x  = strtol (parametersGetInput (params, i + 0), NULL, 0);
      int y  = strtol (parametersGetInput (params, i + 1), NULL, 0);
      int dx = strtol (parametersGetInput (params, i + 2), NULL, 0);
      int dy = strtol (parametersGetInput (params, i + 3), NULL, 0);
      painterDrawLine (self -> painter, x, y, dx, dy);
    }
}

void
canvasCSwapBuffers (struct Canvas *self, struct Parameters *params)
{
  register struct RGBABuffer *t = self -> front;
  painterDestroy (self -> painter);
  self -> front = self -> back;
  self -> back = t;
  self -> painter = rgbabufferGetPainter (self -> back);
  widgetRerender (canvasToWidget (self), NULL);
}

void
canvasCRequestSize (struct Canvas *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 2)
    {
      self -> widget.reqWidth = strtol (parametersGetInput (params, 0), NULL, 0);
      self -> widget.reqHeight = strtol (parametersGetInput (params, 1), NULL, 0);
      widgetReconfigure (canvasToWidget (self));
    }
}

/* arch-tag: 47bb0c50-39ce-4ff0-aca8-138b0ce5922f
 */
