#include <Y/widget/console.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/buffer/painter.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>

#include <Y/text/font.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

/*
 * This whole file, and the console/terminal system in general, needs
 * a serious overhaul.  It was written a bit hastily.  Volunteers?
 */

enum ConsoleCharFlags
{
  CCF_BOLD       = 1<<0,
  CCF_BLINK      = 1<<1,
  CCF_INVERSE    = 1<<2,
  CCF_UNDERLINE  = 1<<3
};

struct ConsoleChar
{
  wchar_t character;
  int8_t foreground;
  int8_t background;
  uint8_t flags;
  uint8_t charset;
};

static struct ConsoleChar consolecharEmpty = { L'\0', -1, -1, 0, 'A' };

struct Console
{
  struct Widget widget;
  int cols, rows;
  int cursorCol, cursorRow;
  struct ConsoleChar *contents;
  uint32_t colours[16];
  int foreground, background, bold, blink, inverse, underline;
  int defaultForeground, defaultBackground;
  int charWidth, charHeight, charBase;
  char charset; 
};

static void consoleResize (struct Widget *);
static void consoleReconfigure (struct Widget *);
static void consolePaint (struct Widget *, struct Painter *);
static int consoleKeyboardRaw (struct Widget *, enum YKeyCode, int, int);

static struct Object *consoleInstantiate (struct Parameters *);

static void consoleCDrawText        (struct Console *, struct Parameters *);
static void consoleCClearRect       (struct Console *, struct Parameters *);
static void consoleCSetRendition    (struct Console *, struct Parameters *);
static void consoleCSwapVideo       (struct Console *, struct Parameters *);
static void consoleCRing            (struct Console *, struct Parameters *);
static void consoleCUpdateCursorPos (struct Console *, struct Parameters *);
static void consoleCScrollView      (struct Console *, struct Parameters *);

CLASSDEF struct Class consoleClass =
{
  name:          "Console",
  super:         &widgetClass,
  instantiate:   consoleInstantiate,
  destroy:       (void (*)(struct Object *))consoleDestroy,
  methodCount:   7,
  methods:
    {
      { "clearRect",       (ObjectMethod_t) consoleCClearRect       },
      { "drawText",        (ObjectMethod_t) consoleCDrawText        },
      { "ring",            (ObjectMethod_t) consoleCRing            },
      { "scrollView",      (ObjectMethod_t) consoleCScrollView      },
      { "setRendition",    (ObjectMethod_t) consoleCSetRendition    },
      { "swapVideo",       (ObjectMethod_t) consoleCSwapVideo       },
      { "updateCursorPos", (ObjectMethod_t) consoleCUpdateCursorPos }
    }
};

static struct WidgetTable consoleTable =
{
  c:             &consoleClass,
  resize:        consoleResize,
  reconfigure:   consoleReconfigure,
  paint:         consolePaint,
  keyboardRaw:   consoleKeyboardRaw,
};

static inline struct Console *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == consoleClass ); */
  return (struct Console *)widget;
}

static inline const struct Console *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == consoleClass ); */
  return (const struct Console *)widget;
}

static void
consoleResizeContents (struct Console *self, int newCols, int newRows)
{
  struct ConsoleChar *newData, *dst, *src;
  int i, j;
  if (newCols == self -> cols && newRows == self -> rows)
    return;

  newData = ymalloc (sizeof (struct ConsoleChar) * newCols * newRows);

  dst = newData;
  for (j=0; j<newRows; ++j)
    {
      src = self -> contents + self -> cols * j;
      for (i=0; i<newCols; ++i)
        {
          if (j < self -> rows && i < self -> cols)
            memcpy (dst, src, sizeof (struct ConsoleChar));
          else
            memcpy (dst, &consolecharEmpty, sizeof (struct ConsoleChar));
          ++src;
          ++dst;
        }
    }

  yfree (self -> contents);
  self -> contents = newData;
  self -> rows = newRows;
  self -> cols = newCols;
  widgetRepaint (consoleToWidget (self), NULL);
}

static void
consoleRepaintChars (struct Console *self, int sCol, int sRow,
                     int eCol, int eRow)
{
  struct Rectangle *dirty;
  dirty = rectangleCreate (sCol * self -> charWidth, sRow * self -> charHeight,
                           (eCol - sCol) * self -> charWidth,
                           (eRow - sRow) * self -> charHeight);
          widgetRepaint (consoleToWidget (self), dirty);
  
}

static void
consoleDrawText (struct Console *self, int col, int row, wchar_t *string,
                 int width)
{
  int i = col, j = row;
  wchar_t *cs = string;
  struct ConsoleChar *cur = self -> contents + self -> cols * j + i;
  int start = i;

  if (i >= self -> cols || j >= self -> rows)
    return;

  while (*cs != L'\0')
    {
      cur -> character = *cs;
      cur -> foreground = self -> foreground;
      cur -> background = self -> background;
      cur -> flags = ((self -> bold)      ? CCF_BOLD      : 0)
                   | ((self -> blink)     ? CCF_BLINK     : 0)
                   | ((self -> inverse)   ? CCF_INVERSE   : 0)
                   | ((self -> underline) ? CCF_UNDERLINE : 0);
      cur -> charset = self -> charset;
      ++i;
      ++cur;
      if (i >= self -> cols)
        {
          consoleRepaintChars (self, start, j, self -> cols, j + 1);
          i = 0;
          start = 0;
          ++j;
          cur = self -> contents + self -> rows * j + i;
          if (j >= self -> rows)
            return; /* scroll? */
        }
      ++cs;
    }
  consoleRepaintChars (self, start, j, i, j + 1);
}

static void
consoleClearRect (struct Console *self, int sCol, int sRow, int eCol, int eRow)
{
  int i, j;
  if (sCol >= self -> cols)
    return;
  if (sRow >= self -> rows)
    return;
  if (eCol > self -> cols)
    eCol = self -> cols;
  if (eRow > self -> rows)
    eRow = self -> rows;
  for (j=sRow; j < eRow; ++j)
    {
      struct ConsoleChar *cur = self -> contents + self -> cols * j + sCol;
      for (i=sCol; i < eCol; ++i)
        {
          cur -> character = L'\0';
          cur -> foreground = -1;
          cur -> background = -1;
          cur -> flags = 0;
          cur -> charset = self -> charset; 
          ++cur;
        }
    }
  consoleRepaintChars (self, sCol, sRow, eCol, eRow);
}

static void
consolePaint (struct Widget *self_w, struct Painter *painter)
{
  struct Console *self = castBack (self_w);
  int i = 0, j = 0;
  struct ConsoleChar *cur;
  struct Font *font = fontCreate ("Andale Mono", "Regular", 11);
  struct Rectangle *painterClip = painterGetClipRectangle (painter);

  int left = painterClip -> x / self -> charWidth;
  int right = (painterClip -> x + painterClip -> w + self -> charWidth - 1)
              / self -> charWidth;
  int top = painterClip -> y / self -> charHeight; 
  int bottom = (painterClip -> y + painterClip -> h + self -> charHeight - 1)
              / self -> charHeight;

  rectangleDestroy (painterClip);

  if (right > self -> cols)
    right = self -> cols;

  if (bottom > self -> rows)
    bottom = self -> rows;

  for (j = top; j < bottom; ++j)
    {
      cur = self -> contents + self -> cols * j + left;
      for (i = left; i < right; ++i)
        {
          int8_t background = cur -> background;
          int8_t foreground = cur -> foreground;
          if (background < 0)
            background = self -> defaultBackground;
          if (foreground < 0)
             foreground = self -> defaultForeground;  
          if ((cur -> flags & CCF_INVERSE) != 0)
            {
              int8_t tmp = background;
              background = foreground;
              foreground = tmp;
            } 
          if (cur -> flags & CCF_BOLD)
            foreground += 8;
          if (cur -> flags & CCF_BLINK)
            background += 8;
          painterSetPenColour  (painter, self -> colours[background % 16]);
          painterSetFillColour (painter, self -> colours[background % 16]);
          painterDrawRectangle (painter, self -> charWidth * i,
                                self -> charHeight * j,
                                self -> charWidth, self -> charHeight);
          if (cur -> character != L'\0')
            {
              wchar_t str[2] = { cur -> character, L'\0' };
              painterSetPenColour (painter, self -> colours[foreground % 16]);
              fontRenderWCString (font, painter, str,
                                  self -> charWidth * i,
                                  self -> charHeight * j + self -> charBase);
            }
          ++cur;
        }
    } 

  if (self -> cols * self -> charWidth < self -> widget.w)
    {
      painterSetPenColour (painter, self -> colours[self -> defaultBackground]);
      painterSetFillColour (painter, self -> colours[self -> defaultBackground]);
      painterDrawRectangle (painter, self -> cols * self -> charWidth, 0,
                            self -> widget.w - self -> cols * self -> charWidth,
                            self -> widget.h);
    }

  if (self -> rows * self -> charHeight < self -> widget.h)
    {
      painterSetPenColour (painter, self -> colours[self -> defaultBackground]);
      painterSetFillColour (painter, self -> colours[self -> defaultBackground]);
      painterDrawRectangle (painter, 0, self -> rows * self -> charHeight,
                            self -> widget.w,
                            self -> widget.h - self -> rows * self -> charHeight);
    }

  fontDestroy (font);
}

int
consoleKeyboardRaw (struct Widget *self_w, enum YKeyCode key,
                    int pressed, int modifierState)
{
  struct Console *self = castBack (self_w);
  if (pressed)
    {
      struct Parameters *params = parametersCreate ();
      parametersAddOutputInteger (params, key);
      parametersAddOutputInteger (params, modifierState);
      objectEmitSignal (consoleToObject (self), "keyPress", params);
      parametersDestroy (params);
    }
  return 1;
}

static int defaultColours[] =
{
  0xBB000000, 0xFFCC0000, 0xFF00CC00, 0xFFCCCC00,
  0xFF4444CC, 0xFFCC00CC, 0xFF00CCCC, 0xFFCCCCCC,
  0xFF666666, 0xFFFF6666, 0xFF66FF66, 0xFFFFFF66,
  0xFF6666FF, 0xFFFF66FF, 0xFF66FFFF, 0xFFFFFFFF
};

struct Console *
consoleCreate ()
{
  struct Console *self = ymalloc (sizeof (struct Console));
  objectInitialise (&(self -> widget.o), &consoleClass);
  widgetInitialise (&(self -> widget), &consoleTable);
  self -> cols = 0;
  self -> rows = 0;
  self -> cursorCol = 0;
  self -> cursorRow = 0;
  self -> contents = NULL;
  memcpy (self -> colours, defaultColours, sizeof (defaultColours));
  self -> defaultForeground = 7;
  self -> defaultBackground = 0;
  self -> foreground = -1;
  self -> background = -1;
  self -> bold = 0;
  self -> blink = 0;
  self -> inverse = 0;
  self -> underline = 0;
  self -> charset = 'A';
  self -> charWidth = 6;
  self -> charHeight = 10;
  self -> charBase = 8;
  consoleResizeContents (self, 80, 24);
  return self;
}

void
consoleDestroy (struct Console *self)
{
  widgetFinalise (consoleToWidget (self));
  objectFinalise (consoleToObject (self));
  yfree (self);
}

struct Widget *
consoleToWidget (struct Console *self)
{
  return &(self -> widget);
}

struct Object *
consoleToObject (struct Console *self)
{
  return &(self -> widget.o);
}

static struct Object *
consoleInstantiate (struct Parameters *p)
{
  return consoleToObject (consoleCreate ());
}

static void
consoleResize (struct Widget *self_w)
{
  struct Console *self = castBack (self_w);
  int newCols = self -> widget.w / self -> charWidth;
  int newRows = self -> widget.h / self -> charHeight;
  if (newCols < 1)
    newCols = 1;
  if (newRows < 1)
    newRows = 1;
  if (newCols != self -> cols || newRows != self -> rows)
    {
      struct Parameters *params = parametersCreate ();
      consoleResizeContents (self, newCols, newRows); 
      parametersAddOutputInteger (params, self -> cols);
      parametersAddOutputInteger (params, self -> rows);
      objectEmitSignal (consoleToObject (self), "resize", params);
      parametersDestroy (params);
    }
}

static void
consoleReconfigure (struct Widget *self_w)
{
  struct Console *self = castBack (self_w);

  self -> widget.reqWidth  = self -> charWidth * 80;
  self -> widget.reqHeight = self -> charHeight * 24;
  self -> widget.minWidth  = self -> charWidth;
  self -> widget.minHeight = self -> charHeight;

  widgetReconfigure (self -> widget.container);
}

void
consoleCDrawText (struct Console *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 5)
    {
      int col, row, length, width;
      const char *string;
      col = strtol (parametersGetInput (params, 0), NULL, 0);
      row = strtol (parametersGetInput (params, 1), NULL, 0);
      string = parametersGetInput (params, 2);
      length = strtol (parametersGetInput (params, 3), NULL, 0);
      width = strtol (parametersGetInput (params, 4), NULL, 0);
      {
        wchar_t wstring[2 * (length + 1)];
        mbstowcs (wstring, string, 2 * (length + 1));
        consoleDrawText (self, col, row, wstring, width);
      }
    }
}

void
consoleCClearRect (struct Console *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 4)
    {
      int sCol, sRow, eCol, eRow;
      sCol = strtol (parametersGetInput (params, 0), NULL, 0);
      sRow = strtol (parametersGetInput (params, 1), NULL, 0);
      eCol = strtol (parametersGetInput (params, 2), NULL, 0);
      eRow = strtol (parametersGetInput (params, 3), NULL, 0);
      consoleClearRect (self, sCol, sRow, eCol, eRow);
    }
}

void
consoleCSetRendition (struct Console *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 7)
    {
      int bold, blink, inverse, underline, foreground, background;
      const char *charset; 
      bold       = strtol (parametersGetInput (params, 0), NULL, 0);
      blink      = strtol (parametersGetInput (params, 1), NULL, 0);
      inverse    = strtol (parametersGetInput (params, 2), NULL, 0);
      underline  = strtol (parametersGetInput (params, 3), NULL, 0);
      foreground = strtol (parametersGetInput (params, 4), NULL, 0);
      background = strtol (parametersGetInput (params, 5), NULL, 0);
      charset    = parametersGetInput (params, 6);

      self -> bold = bold;
      self -> blink = blink;
      self -> inverse = inverse;
      self -> underline = underline;
      if (foreground >= 1 && foreground <= 8)
        self -> foreground = foreground - 1;
      else
        self -> foreground = -1;
      if (background >= 1 && background <= 8)
        self -> background = background - 1;
      else
        self -> background = -1;
      self -> charset = charset[0];
   }
}

void
consoleCSwapVideo (struct Console *self, struct Parameters *params)
{
  Y_TRACE ("Swap Video ?");
}

void
consoleCRing (struct Console *self, struct Parameters *params)
{
  Y_TRACE ("***BEEP***");
}

void
consoleCUpdateCursorPos (struct Console *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 2)
    {
      int col, row;
      col = strtol (parametersGetInput (params, 0), NULL, 0);
      row = strtol (parametersGetInput (params, 1), NULL, 0);
      if (col != self -> cursorCol || row != self -> cursorRow)
        {
          consoleRepaintChars (self, self -> cursorCol, self -> cursorRow,
                                     self -> cursorCol + 1, self -> cursorRow +1);
          self -> cursorCol = col;
          self -> cursorRow = row;
          consoleRepaintChars (self, self -> cursorCol, self -> cursorRow,
                                     self -> cursorCol + 1, self -> cursorRow +1);
        }
    } 
}

void
consoleCScrollView (struct Console *self, struct Parameters *params)
{
  if (parametersGetInputCount (params) == 3)
    {
      int destRow, srcRow, numLines;
      destRow  = strtol (parametersGetInput (params, 0), NULL, 0);
      srcRow   = strtol (parametersGetInput (params, 1), NULL, 0);
      numLines = strtol (parametersGetInput (params, 2), NULL, 0);
      if (destRow >= self -> rows)
        return;
      if (srcRow >= self -> rows)
        return;
      if (destRow + numLines > self -> rows)
        numLines = self -> rows - destRow;
      if (srcRow + numLines > self -> rows)
        numLines = self -> rows - srcRow;
      memmove (self -> contents + destRow * self -> cols,
               self -> contents + srcRow * self -> cols,
               numLines * self -> cols * sizeof (struct ConsoleChar)); 
      if (destRow < srcRow)
        consoleClearRect (self, 0, MAX (srcRow, destRow + numLines),
                          self -> cols, srcRow + numLines);
      else if (srcRow < destRow)
        consoleClearRect (self, 0, srcRow, self -> cols,
                          MIN (numLines, destRow - srcRow));
      consoleRepaintChars (self, 0, destRow, self -> cols, destRow + numLines);
    }
}

/* arch-tag: 0b2da405-d961-435e-afae-1cf8f01f8310
 */
