#include <Y/widget/gridlayout.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/buffer/painter.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>

#include <Y/text/font.h>

#include <stdio.h>
#include <ctype.h>

#ifndef MAX
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#endif

struct GridLayout
{
  struct Widget widget;
  int rows, cols;
  int *rowHeights, *colWidths;
  struct List *items;
  struct Widget *pointerWidget;
};

struct GridItem
{
  struct Widget *widget;
  int row, col, rowspan, colspan;
};

static void gridlayoutUnpack (struct Widget *, struct Widget *);
static void gridlayoutPaint (struct Widget *, struct Painter *);
static void gridlayoutResize (struct Widget *);
static void gridlayoutReconfigure (struct Widget *);
static int gridlayoutPointerMotion (struct Widget *, int, int, int, int);
static int gridlayoutPointerButton (struct Widget *, int, int, int, int);
static void gridlayoutPointerEnter (struct Widget *, int, int);
static void gridlayoutPointerLeave (struct Widget *);

static void gridlayoutFitChildren (struct GridLayout *self);

static struct Object *gridlayoutInstantiate (struct Parameters *);

static void gridlayoutCAddWidget (struct Object *, struct Parameters *);

CLASSDEF struct Class gridlayoutClass =
{
  name:          "GridLayout",
  super:         &widgetClass,
  instantiate:   gridlayoutInstantiate,
  destroy:       (void (*)(struct Object *))gridlayoutDestroy,
  methodCount:   1,
  methods:
    { { "addWidget", gridlayoutCAddWidget } }
};

static struct WidgetTable gridlayoutTable =
{
  c:             &gridlayoutClass,
  unpack:        gridlayoutUnpack,
  reconfigure:   gridlayoutReconfigure,
  resize:        gridlayoutResize,
  paint:         gridlayoutPaint,
  pointerMotion: gridlayoutPointerMotion,
  pointerButton: gridlayoutPointerButton,
  pointerEnter:  gridlayoutPointerEnter,
  pointerLeave:  gridlayoutPointerLeave
};

static inline struct GridLayout *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == gridlayoutClass ); */
  return (struct GridLayout *)widget;
}

static inline const struct GridLayout *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == gridlayoutClass ); */
  return (const struct GridLayout *)widget;
}

static void
gridlayoutUnpack (struct Widget *self_w, struct Widget *w)
{
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;

  iter = listGetStartIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      if (item -> widget == w)
        {
          listiteratorRemove (iter);
          yfree (item);
          widgetSetContainer (w, NULL);
          listiteratorDestroy (iter);
          return;
        }
      listiteratorNext (iter);
    }
  listiteratorDestroy (iter);
}

static void
gridlayoutPaint (struct Widget *self_w, struct Painter *painter)
{
  /* draw each child */
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;

  iter = listGetStartIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      struct Rectangle *rect = widgetGetRectangle (item->widget);
      painterSaveState (painter);
      painterEnter (painter, rect);
      if (!painterFullyClipped (painter))
        widgetPaint (item -> widget, painter);
      painterRestoreState (painter); 
      rectangleDestroy (rect);
      listiteratorNext (iter);
    }
  listiteratorDestroy (iter);
}

struct GridLayout *
gridlayoutCreate ()
{
  struct GridLayout *self = ymalloc (sizeof (struct GridLayout));
  objectInitialise (&(self -> widget.o), &gridlayoutClass);
  widgetInitialise (&(self -> widget), &gridlayoutTable);
  self -> rows = 0;
  self -> cols = 0;
  self -> rowHeights = NULL;
  self -> colWidths = NULL;
  self -> items = listCreate ();
  self -> pointerWidget = NULL;
  return self;
}

static void
griditemDestroy (struct GridItem *item)
{
  widgetSetContainer (item -> widget, NULL);
  yfree (item);
}

void
gridlayoutDestroy (struct GridLayout *self)
{
  listDestroy (self -> items, (void(*)(void*))griditemDestroy);
  widgetFinalise (gridlayoutToWidget (self));
  objectFinalise (gridlayoutToObject (self));
  yfree (self);
}

struct Widget *
gridlayoutToWidget (struct GridLayout *self)
{
  return &(self -> widget);
}

struct Object *
gridlayoutToObject (struct GridLayout *self)
{
  return &(self -> widget.o);
}

static struct Object *
gridlayoutInstantiate (struct Parameters *p)
{
  return gridlayoutToObject (gridlayoutCreate ());
}

static void
gridlayoutResizeGrid (struct GridLayout *self, int cols, int rows)
{
  struct ListIterator *iter;

  self -> cols = cols;
  self -> rows = rows;

  /* check that we aren't resizing to smaller than any child */
  iter = listGetStartIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      self -> cols = MAX (self -> cols, item -> col + item -> colspan);
      self -> rows = MAX (self -> rows, item -> row + item -> rowspan);
      listiteratorNext (iter);
    }

  listiteratorDestroy (iter);

  yfree (self -> rowHeights);
  yfree (self -> colWidths);
  self -> rowHeights = ymalloc (sizeof (int) * self -> rows);
  self -> colWidths  = ymalloc (sizeof (int) * self -> cols);
  
  gridlayoutFitChildren (self);
}

static int
rowHeight (struct GridLayout *self, int row, int span)
{
  int i, h;
  for (i=0, h=0; i<span; ++i)
    h += self -> rowHeights[row + i];
  return h;
}

static int
colWidth (struct GridLayout *self, int col, int span)
{
  int i, w;
  for (i=0, w=0; i<span; ++i)
    w += self -> colWidths[col + i];
  return w;
}

static void
gridlayoutFitChildren (struct GridLayout *self)
{
  /* homogeneous, for now */
  struct ListIterator *iter;
  int i, j;
  int cumulativeWidth, cumulativeHeight;

  /* work out column widths and row heights */
  cumulativeWidth = 0;
  for (i = 0; i < self->cols; ++i)
    {
      self->colWidths[i] = (self->widget.w - cumulativeWidth) /
                           (self -> cols - i);
      cumulativeWidth += self->colWidths[i];
    }

  cumulativeHeight = 0;
  for (j = 0; j < self->rows; ++j)
    {
      self->rowHeights[j] = (self->widget.h - cumulativeHeight) /
                            (self -> rows - j);
      cumulativeHeight += self->rowHeights[j];
    }

  iter = listGetStartIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      struct Rectangle *childRect = widgetGetRectangle (item -> widget);
      int x, y, w, h;
      x = colWidth (self, 0, item -> col);
      y = rowHeight (self, 0, item -> row);
      w = colWidth (self, item -> col, item -> colspan);
      h = rowHeight (self, item -> row, item -> rowspan);
      if (childRect -> x != x || childRect -> y != y)
        widgetMove (item -> widget, x, y);
      if (childRect -> w != w || childRect -> h != h)
        widgetResize (item -> widget, w, h);
      rectangleDestroy (childRect);
      listiteratorNext (iter);
    }
  listiteratorDestroy (iter);
}

static void
gridlayoutCAddWidget (struct Object *self_o, struct Parameters *params)
{
  struct GridLayout *self = (struct GridLayout *)self_o;
  if (self != NULL || parametersGetInputCount (params) > 4)
    {
      struct GridItem *item;

      item = ymalloc (sizeof (struct GridItem));    
      item -> widget = (struct Widget *)
            objectFind (strtol (parametersGetInput (params, 0), NULL, 0));
      item -> col     = strtol (parametersGetInput (params, 1), NULL, 0);
      item -> row     = strtol (parametersGetInput (params, 2), NULL, 0);
      item -> colspan = strtol (parametersGetInput (params, 3), NULL, 0);
      item -> rowspan = strtol (parametersGetInput (params, 4), NULL, 0);

      listAppend (self -> items, item);

      if (item -> col + item -> colspan > self -> cols ||
          item -> row + item -> rowspan > self -> rows)
        {
           gridlayoutResizeGrid (self, self -> cols, self -> rows);
        }
      else
        {
           gridlayoutFitChildren (self);
           widgetRepaint (gridlayoutToWidget (self),
                          widgetGetRectangle (item -> widget));
        }
      widgetReconfigure (gridlayoutToWidget (self));
      widgetSetContainer (item -> widget, gridlayoutToWidget (self));
    }
}

void
gridlayoutReconfigure (struct Widget *self_w)
{
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;
  int minRowHeight = -1, minColWidth = -1;
  int reqRowHeight = -1, reqColWidth = -1;

  /* homogeneous */

  iter = listGetStartIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      int mcw = item -> widget -> minWidth / item -> colspan;
      int mrh = item -> widget -> minHeight / item -> rowspan;
      int rcw = item -> widget -> reqWidth / item -> colspan;
      int rrh = item -> widget -> reqHeight / item -> rowspan;
      if (mcw > minColWidth)
        minColWidth = mcw;
      if (mrh > minRowHeight)
        minRowHeight = mrh;
      if (rcw > reqColWidth)
        reqColWidth = rcw;
      if (rrh > reqRowHeight)
        reqRowHeight = rrh;
      listiteratorNext (iter);
    }
  listiteratorDestroy (iter);

  if (minColWidth > 0)
    self -> widget.minWidth = minColWidth * self -> cols;
  else
    self -> widget.minWidth = -1;
  if (minRowHeight > 0)
    self -> widget.minHeight = minRowHeight * self -> rows;
  else
    self -> widget.minHeight = -1;

  if (reqColWidth < minColWidth)
    reqColWidth = minColWidth;
  if (reqRowHeight < minRowHeight)
    reqRowHeight = minRowHeight;

  if (reqColWidth > 0)
    self -> widget.reqWidth = reqColWidth * self -> cols;
  else
    self -> widget.reqWidth = -1;
  if (reqRowHeight > 0)
    self -> widget.reqHeight = reqRowHeight * self -> rows;
  else
    self -> widget.reqHeight = -1;

  widgetReconfigure (self -> widget.container);
}

void
gridlayoutResize (struct Widget *self_w)
{
  struct GridLayout *self = castBack (self_w);
  gridlayoutFitChildren (self); 
}

int
gridlayoutPointerMotion (struct Widget *self_w, int x, int y, int dx, int dy)
{
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;

  /* iterate backwards, so items that appear on top get served first */
  iter = listGetEndIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      int lx = x - item -> widget -> x;
      int ly = y - item -> widget -> y;
      if (widgetContainsPoint (item -> widget, lx, ly))
        {
          if (self -> pointerWidget != item -> widget)
            {
              if (self -> pointerWidget != NULL)
                widgetPointerLeave (self -> pointerWidget);
              self -> pointerWidget = item -> widget;
              widgetPointerEnter (item -> widget, lx, ly);
            }
          widgetPointerMotion (item -> widget, x, y, dx, dy);
          listiteratorDestroy (iter);
          return 1;
        } 
      listiteratorPrevious (iter);
    }
  listiteratorDestroy (iter);

  if (self -> pointerWidget != NULL)
    {
      widgetPointerLeave (self -> pointerWidget);
      self -> pointerWidget = NULL;
    }
  return 0;
}

int
gridlayoutPointerButton (struct Widget *self_w, int x, int y, int b, int p)
{
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;

  /* iterate backwards, so items that appear on top get served first */
  iter = listGetEndIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      int lx = x - item -> widget -> x;
      int ly = y - item -> widget -> y;
      if (widgetContainsPoint (item -> widget, lx, ly)
          && widgetPointerButton (item -> widget, lx, ly, b, p))
        {
          listiteratorDestroy (iter);
          return 1;
        } 
      listiteratorPrevious (iter);
    }
  listiteratorDestroy (iter);
  return 0;
}

void
gridlayoutPointerEnter (struct Widget *self_w, int x, int y)
{
  struct GridLayout *self = castBack (self_w);
  struct ListIterator *iter;

  /* iterate backwards, so items that appear on top get served first */
  iter = listGetEndIterator (self -> items);
  while (listiteratorHasValue (iter))
    {
      struct GridItem *item = listiteratorGet (iter);
      int lx = x - item -> widget -> x;
      int ly = y - item -> widget -> y;
      if (widgetContainsPoint (item -> widget, lx, ly))
        {
          self -> pointerWidget = item -> widget;
          widgetPointerEnter (item -> widget, lx, ly);
          listiteratorDestroy (iter);
          return;
        } 
      listiteratorPrevious (iter);
    }
  listiteratorDestroy (iter);
  return;
}

void
gridlayoutPointerLeave (struct Widget *self_w)
{
  struct GridLayout *self = castBack (self_w);

  if (self -> pointerWidget != NULL)
    {
      widgetPointerLeave (self -> pointerWidget);
      self -> pointerWidget = NULL;
    }
}


/* arch-tag: 24031e9a-1ee1-4239-aa61-d876e524b995
 */
