#include <Y/modules/module_interface.h>

#include <Y/message/message.h>
#include <Y/message/client.h>
#include <Y/message/client_p.h>

#include <Y/util/yutil.h>
#include <Y/util/list.h>
#include <Y/util/index.h>

#include <Y/main/control.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <errno.h>

#define MESSAGE_HEADER_SIZE (4 * 8)

struct TcpData
{
  int port;
  int listeningFD;
  struct Index *tcpClients;
};

struct TcpClient
{
  struct Client client;
  int fd;
  struct TcpData *moduleData;
  struct Message *sending;
  struct Message *receiving;
  ssize_t sent, received;
  struct List *sendQueue;
};

static void tcpSendMessage (struct Client *c, struct Message *m);
static void tcpClose (struct Client *c);

struct ClientClass tcpClientClass =
{
  name: "TCP Socket Client",
  sendMessage: tcpSendMessage,
  close: tcpClose
};

static struct TcpClient *
castBack (struct Client *self_c)
{
  /* assert ( self_c -> c == &tcpClientClass ); */
  return (struct TcpClient *)self_c;
}

static void
tcpSendMessage (struct Client *self_c, struct Message *m)
{
  struct TcpClient *self = castBack (self_c);
  if (self -> sending == NULL)
    {
      self -> sending = m;
      self -> sent = 0;
      controlChangeFileDescriptorMask (self -> fd,
                 CONTROL_WATCH_WRITE | CONTROL_WATCH_READ );
    }
  else
    {
      listAppend (self -> sendQueue, m);
    }
}

static void
tcpClose (struct Client *self_c)
{
  struct TcpClient *self = castBack (self_c);
  indexRemove (self -> moduleData -> tcpClients, &(self -> client.id));
  controlUnregisterFileDescriptor (self -> fd);
  close (self -> fd);
  listDestroy (self -> sendQueue, (void (*)(void *))messageDestroy);
  messageDestroy (self -> sending);
  messageDestroy (self -> receiving);
  yfree (self);
}



static void
tcpClientReady (int fd, int causeMask, void *data_v)
{
  struct TcpClient *self = data_v;
  if (causeMask & CONTROL_WATCH_READ)
    {
      if (self -> receiving == NULL)
        {
          self -> receiving = ymalloc (sizeof (struct Message));
          self -> receiving -> data = ymalloc (MESSAGE_HEADER_SIZE);
          self -> received = 0;
        }
      if (self -> received < MESSAGE_HEADER_SIZE)
        {
          /* attempt to read the rest of the header */
          ssize_t ret = read (self -> fd,
                              self -> receiving -> data + self -> received,
                              MESSAGE_HEADER_SIZE - self -> received);
          if (ret < 1)
            {
              clientClose (&(self -> client));
              return;
            }
          self -> received += ret;
          if (self -> received == MESSAGE_HEADER_SIZE)
            {
              /* header fully received, translate */
              uint32_t *netdata = (uint32_t *)self -> receiving -> data;

              self -> receiving -> to   = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> from = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> cid  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> oid  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> op   = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> meta = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> ext  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> len  = (int) ntohl (*netdata); ++netdata;

              /* enforce client ID onto message from (to avoid spoofing) */
              self -> receiving -> from = self -> client.id;

              yfree (self -> receiving -> data);

              if (self -> receiving -> len == 0)
                {
                  /* got the (bodyless) message */
                  self -> receiving -> data = NULL;
                  messageDespatch (&(self -> client), self -> receiving);
                  self -> receiving = NULL;
                }
              else
                {
                  self -> receiving -> data =
                         ymalloc (self -> receiving -> len + 1);
                  self -> receiving -> data[self -> receiving -> len] = '\0';
                }
            }
        }
      else if (self -> received >= MESSAGE_HEADER_SIZE)
        {
          ssize_t ret = read (self -> fd, self -> receiving -> data
                                  + self -> received - MESSAGE_HEADER_SIZE,
                              self -> receiving -> len - (self -> received
                                                    - MESSAGE_HEADER_SIZE));
          if (ret < 1)
            {
              clientClose (&(self -> client));
              return;
            }

          self -> received += ret;
          if (self -> received == self -> receiving -> len + MESSAGE_HEADER_SIZE)
            {
              /* got the message */
              messageDespatch (&(self -> client), self -> receiving);
              self -> receiving = NULL;
            }
        }
    }
  if (causeMask & CONTROL_WATCH_WRITE)
    {
      if (self -> sent < MESSAGE_HEADER_SIZE)
        {
          /* attempt to write the rest of the header */
          ssize_t ret;
          char data[MESSAGE_HEADER_SIZE];
          uint32_t *netdata = (uint32_t *)data;
          struct iovec buffers[2];

          (*(netdata++)) = htonl ((uint32_t) self -> sending -> to);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> from);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> cid);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> oid);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> op);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> meta);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> ext);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> len);

          buffers[0].iov_base = data + self -> sent;
          buffers[0].iov_len = MESSAGE_HEADER_SIZE - self -> sent;
          buffers[1].iov_base = self -> sending -> data;
          buffers[1].iov_len = self -> sending -> len;

          ret = writev (self -> fd, buffers, self -> sending -> len ? 2 : 1); 

          if (ret == 0)
            {
              perror ("whilst writing header");
              Y_FATAL ("Cannot write anything");
            }

          self -> sent += ret;
        }
      else 
        {
          ssize_t ret;
          ret = write (self -> fd, self -> sending -> data + (self -> sent - MESSAGE_HEADER_SIZE),
                       self -> sending -> len - (self -> sent - MESSAGE_HEADER_SIZE));
          if (ret == 0)
            {
              perror ("whilst writing body");
              Y_FATAL ("Cannot write anything");
            }

          self -> sent += ret;
        }
      if (self -> sent == self -> sending -> len + MESSAGE_HEADER_SIZE)
        {
          /* message sent */
          messageDestroy (self -> sending);
          if (listCount (self -> sendQueue) > 0)
            { 
              self -> sent = 0;
              self -> sending = listPopFront (self -> sendQueue);
            }
          else
            {
              self -> sending = NULL;
              controlChangeFileDescriptorMask (self -> fd,
                                               CONTROL_WATCH_READ );
            }
        }
    }
}

static void
tcpSocketReady (int fd, int causeMask, void *data_v)
{
  struct TcpData *data = data_v;
  struct TcpClient *newClient;

  newClient = ymalloc (sizeof (struct TcpClient));
  newClient -> fd = accept (fd, NULL, NULL);
  newClient -> moduleData = data;
  newClient -> sending = NULL;
  newClient -> receiving = NULL;
  newClient -> sent = 0;
  newClient -> received = 0;
  newClient -> sendQueue = listCreate ();

  newClient -> client.c = &tcpClientClass;

  indexAdd (data -> tcpClients, newClient);

  clientRegister (&(newClient -> client));

  controlRegisterFileDescriptor (newClient -> fd, CONTROL_WATCH_READ,
                                 newClient, tcpClientReady);
  
};

int
initialise (struct Module *self, int argc, const char **argv)
{
  struct TcpData *data = ymalloc (sizeof (struct TcpData));
  struct sockaddr_in sockaddr;
  int i, r;

  self -> name = "IPC: TCP Sockets";
  self -> data = data;
  data -> port = 8900;

  for (i=0; i<argc; ++i)
    {
      if (strncmp (argv[i], "port=", 5) == 0
          && strlen (argv[i]+5) > 0)
        data -> port = strtoul (argv[i]+5, NULL, 10);
    }

  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons (data -> port);
  sockaddr.sin_addr.s_addr = INADDR_ANY;

  data -> listeningFD = socket (PF_INET, SOCK_STREAM, 0);

  if (data -> listeningFD < 0)
    {
      /* clean up */
      yfree (data);
      Y_ERROR ("Could not open socket: %s", strerror (errno));
      return 1;
    }
 
  r = bind (data->listeningFD, (struct sockaddr *)&sockaddr, sizeof (sockaddr));

  if (r < 0)
    {
      /* clean up */
      close (data -> listeningFD);
      yfree (data);
      Y_ERROR ("Could not bind to socket: %s", strerror (errno));
      return 1;
    }

  r = listen (data->listeningFD, 64);

  if (r < 0)
    {
      /* clean up */
      close (data -> listeningFD);
      yfree (data);
      Y_ERROR ("Could not listen to socket: %s", strerror (errno));
      return 1;
    }

  data -> tcpClients = indexCreate (clientKeyFunction,
                                    clientComparisonFunction);
    
  controlRegisterFileDescriptor (data->listeningFD, CONTROL_WATCH_READ,
                                 data, tcpSocketReady);

  return 0;
}

int
finalise (struct Module *self)
{
  struct TcpData *data = self -> data;
  indexDestroy (data -> tcpClients, (void (*)(void *)) clientClose);
  controlUnregisterFileDescriptor (data -> listeningFD);
  close (data -> listeningFD);
  yfree (data);
  return 0;
}

/* arch-tag: 215b4026-8d84-4099-8d28-bcd810df0fe9
 */
